/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Jim Mathies <jmathies@mozilla.com>.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */


#include <windows.h>
#include <stdio.h>
#include <gdiplus.h>
using namespace Gdiplus;

#pragma comment(lib, "gdiplus.lib")

#include "plugin.h"

void SetSubclass(HWND hWnd, NPWindow* instanceData);
void ClearSubclass(HWND hWnd);

#ifndef WM_TOUCHMOVE

/*
 * Touch Input defines and functions
 */

#define WM_TOUCHMOVE                    0x0240
#define WM_TOUCHDOWN                    0x0241
#define WM_TOUCHUP                      0x0242

/*
 * Touch input flag values (TOUCHINPUT.dwFlags)
 */
#define TOUCHEVENTF_MOVE            0x0001
#define TOUCHEVENTF_DOWN            0x0002
#define TOUCHEVENTF_UP              0x0004
#define TOUCHEVENTF_INRANGE         0x0008
#define TOUCHEVENTF_PRIMARY         0x0010
#define TOUCHEVENTF_NOCOALESCE      0x0020
#define TOUCHEVENTF_PEN             0x0040

/*
 * Touch input mask values (TOUCHINPUT.dwMask)
 */
#define TOUCHINPUTMASKF_TIMEFROMSYSTEM  0x0001  // the dwTime field contains a system generated value
#define TOUCHINPUTMASKF_EXTRAINFO       0x0002  // the dwExtraInfo field is valid
#define TOUCHINPUTMASKF_CONTACTAREA     0x0004  // the cxContact and cyContact fields are valid

#endif

typedef BOOL (WINAPI * RegisterTouchWindowPtr)(HWND hwnd, ULONG ulFlags);
static RegisterTouchWindowPtr registerTouchWindow;

typedef BOOL (WINAPI * UnregisterTouchWindowPtr)(HWND hwnd);
static UnregisterTouchWindowPtr unregisterTouchWindow;

typedef BOOL (WINAPI * GetTouchInputInfoPtr)(HANDLE hTouchInput, UINT cInputs, PTOUCHINPUT pInputs, int cbSize);
static GetTouchInputInfoPtr getTouchInputInfo;

typedef BOOL (WINAPI * CloseTouchInputHandlePtr)(HANDLE hTouchInput);
static CloseTouchInputHandlePtr closeTouchInputHandle;

typedef BOOL (WINAPI * IsTouchWindowPtr)(HWND hwnd, ULONG * ulFlags);
static IsTouchWindowPtr isTouchWindow;

struct Junk {
  ULONG_PTR gdiplusToken;
  HWND hWnd;
};

NPNetscapeFuncs NPNFuncs;

/* exports */

NPError OSCALL 
NP_GetEntryPoints(NPPluginFuncs* pFuncs)
{
  pFuncs->version = 11;
  pFuncs->newp = NPP_New;
  pFuncs->destroy = NPP_Destroy;
  pFuncs->setwindow = NPP_SetWindow;
  pFuncs->newstream = NPP_NewStream;
  pFuncs->destroystream = NPP_DestroyStream;
  pFuncs->asfile = NPP_StreamAsFile;
  pFuncs->writeready = NPP_WriteReady;
  pFuncs->write = NPP_Write;
  pFuncs->print = NPP_Print;
  pFuncs->event = NPP_HandleEvent;
  pFuncs->urlnotify = NPP_URLNotify;
  pFuncs->getvalue = NPP_GetValue;
  pFuncs->setvalue = NPP_SetValue;
  pFuncs->javaClass = NULL;

  return NPERR_NO_ERROR;
}

NPError OSCALL 
NP_Initialize(NPNetscapeFuncs* pFuncs)
{
  if(pFuncs == NULL)
    return NPERR_INVALID_FUNCTABLE_ERROR;

  if(HIBYTE(pFuncs->version) > NP_VERSION_MAJOR)
    return NPERR_INCOMPATIBLE_VERSION_ERROR;

  if(pFuncs->size < sizeof(NPNetscapeFuncs))
    return NPERR_INVALID_FUNCTABLE_ERROR;

  NPNFuncs.size                    = pFuncs->size;
  NPNFuncs.version                 = pFuncs->version;
  NPNFuncs.geturlnotify            = pFuncs->geturlnotify;
  NPNFuncs.geturl                  = pFuncs->geturl;
  NPNFuncs.posturlnotify           = pFuncs->posturlnotify;
  NPNFuncs.posturl                 = pFuncs->posturl;
  NPNFuncs.requestread             = pFuncs->requestread;
  NPNFuncs.newstream               = pFuncs->newstream;
  NPNFuncs.write                   = pFuncs->write;
  NPNFuncs.destroystream           = pFuncs->destroystream;
  NPNFuncs.status                  = pFuncs->status;
  NPNFuncs.uagent                  = pFuncs->uagent;
  NPNFuncs.memalloc                = pFuncs->memalloc;
  NPNFuncs.memfree                 = pFuncs->memfree;
  NPNFuncs.memflush                = pFuncs->memflush;
  NPNFuncs.reloadplugins           = pFuncs->reloadplugins;
  NPNFuncs.getJavaEnv              = pFuncs->getJavaEnv;
  NPNFuncs.getJavaPeer             = pFuncs->getJavaPeer;
  NPNFuncs.getvalue                = pFuncs->getvalue;
  NPNFuncs.setvalue                = pFuncs->setvalue;
  NPNFuncs.invalidaterect          = pFuncs->invalidaterect;
  NPNFuncs.invalidateregion        = pFuncs->invalidateregion;
  NPNFuncs.forceredraw             = pFuncs->forceredraw;
  NPNFuncs.getstringidentifier     = pFuncs->getstringidentifier;
  NPNFuncs.getstringidentifiers    = pFuncs->getstringidentifiers;
  NPNFuncs.getintidentifier        = pFuncs->getintidentifier;
  NPNFuncs.identifierisstring      = pFuncs->identifierisstring;
  NPNFuncs.utf8fromidentifier      = pFuncs->utf8fromidentifier;
  NPNFuncs.intfromidentifier       = pFuncs->intfromidentifier;
  NPNFuncs.createobject            = pFuncs->createobject;
  NPNFuncs.retainobject            = pFuncs->retainobject;
  NPNFuncs.releaseobject           = pFuncs->releaseobject;
  NPNFuncs.invoke                  = pFuncs->invoke;
  NPNFuncs.invokeDefault           = pFuncs->invokeDefault;
  NPNFuncs.evaluate                = pFuncs->evaluate;
  NPNFuncs.getproperty             = pFuncs->getproperty;
  NPNFuncs.setproperty             = pFuncs->setproperty;
  NPNFuncs.removeproperty          = pFuncs->removeproperty;
  NPNFuncs.hasproperty             = pFuncs->hasproperty;
  NPNFuncs.hasmethod               = pFuncs->hasmethod;
  NPNFuncs.releasevariantvalue     = pFuncs->releasevariantvalue;
  NPNFuncs.setexception            = pFuncs->setexception;
  return NPERR_NO_ERROR;
}

NPError OSCALL 
NP_Shutdown(void)
{
  return NPERR_NO_ERROR;
}

const char*
NP_GetMIMEDescription(void)
{
  return "application/basic-plugin:basic plugin";
}

/* npp calls */

NPError NPP_New(NPMIMEType pluginType, NPP instance, uint16_t mode, int16_t argc, char* argn[], char* argv[], NPSavedData* saved)
{ 
  HMODULE hLibraryHandle = ::LoadLibraryW(L"user32.dll");

  registerTouchWindow = (RegisterTouchWindowPtr)GetProcAddress(hLibraryHandle, "RegisterTouchWindow");
  unregisterTouchWindow = (UnregisterTouchWindowPtr)GetProcAddress(hLibraryHandle, "UnregisterTouchWindow");
  getTouchInputInfo = (GetTouchInputInfoPtr)GetProcAddress(hLibraryHandle, "GetTouchInputInfo");
  closeTouchInputHandle = (CloseTouchInputHandlePtr)GetProcAddress(hLibraryHandle, "CloseTouchInputHandle");
  isTouchWindow = (IsTouchWindowPtr) GetProcAddress(hLibraryHandle, "IsTouchWindow");
  
  Junk * junk = new Junk;
  if (!junk)
  return NPERR_OUT_OF_MEMORY_ERROR;
  
  memset(junk, 0, sizeof(Junk));
  
  // Initialize GDI+.
  GdiplusStartupInput gdiplusStartupInput;
  GdiplusStartup(&junk->gdiplusToken, &gdiplusStartupInput, NULL);
  
  instance->pdata = (void*)junk;

  BOOL requestChanges = TRUE;
  bool success = (NPERR_NO_ERROR == NPN_SetValue(instance, NPPVpluginRequiresAudioDeviceChanges, (void*)&requestChanges));
  printf("NPAUDIOTEST: "
	  "Register for default audio events: %s\n", success ? "succeess" : "failure");

  return NPERR_NO_ERROR;
}

NPError
NPP_Destroy(NPP instance, NPSavedData** save)
{
  Junk * junk = (Junk*)instance->pdata;
  if (!junk)
    return NPERR_NO_ERROR;
  
  ClearSubclass(junk->hWnd);
  
  // Shutdown GDI+
  GdiplusShutdown(junk->gdiplusToken);
  delete junk;
  instance->pdata = 0;
  return NPERR_NO_ERROR;
}

NPError
NPP_SetWindow(NPP instance, NPWindow* window)
{
  HWND hWnd = (HWND)window->window;

  Junk * junk = (Junk*)instance->pdata;
  if (!junk)
    return NPERR_GENERIC_ERROR;

  if (junk->hWnd == hWnd)
    return NPERR_NO_ERROR;
    
  ClearSubclass(junk->hWnd);

  DWORD dwRes = registerTouchWindow(hWnd, 0);
  if (dwRes == 0) {
    int i = 0 + 1;
  }

  ULONG flags = 0;
  BOOL isTouch = isTouchWindow(hWnd, &flags);
  
  junk->hWnd = hWnd;

  SetSubclass(hWnd, window);

  isTouch = isTouchWindow(hWnd, &flags);

  return NPERR_NO_ERROR;
}

/* windowed plugin events */

void Paint(NPWindow* window)
{
  HDC hdc = NULL;
  PAINTSTRUCT ps;

  hdc = ::BeginPaint((HWND)window->window, &ps);

  if (hdc == NULL)
    return;

  // Reset the clipping region
  SelectClipRgn(hdc, NULL);

  int x = 0;
  int y = 0;
  int width = window->width - 1;
  int height = window->height - 1;

  // Calculate the rectangle coordinates from the instanceData information
  Rect rect(x, y, width, height);

  // Create a layout rect for the text
  RectF boundRect((float)x, (float)y, (float)width, (float)height);
  boundRect.Inflate(-10.0, -10.0);

  Graphics graphics(hdc);

  // Fill in the background and border
  Pen blackPen(Color(255, 0, 0, 0), 1);
  SolidBrush grayBrush(Color(255, 192, 192, 192));

  graphics.FillRectangle(&grayBrush, rect);
  graphics.DrawRectangle(&blackPen, rect);

  ::EndPaint((HWND)window->window, &ps);
}

LRESULT CALLBACK PluginWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg >= 0x240 && uMsg <= 0x242) {
    printf("PluginWndProc: hwnd=0x%X msg=0x%X\n", (unsigned int)hWnd, uMsg);
  }

  ULONG flags = 0;
  BOOL isTouch = isTouchWindow(hWnd, &flags);
//  printf("IsTouch=%d\n", isTouch);
  
  WNDPROC wndProc = (WNDPROC)GetProp(hWnd, TEXT("MozillaWndProc"));
  if (!wndProc)
    return 0;

  NPWindow* window = (NPWindow*)GetProp(hWnd, TEXT("MozillaWindow"));
  if (!window)
    return 0;

  if (uMsg == WM_PAINT) {
    Paint(window);
    return 0;
  }
  if (uMsg == WM_CLOSE) {
    ClearSubclass(hWnd);
  }

  switch(uMsg)
  {
      case WM_TOUCHDOWN:
      case WM_TOUCHMOVE:
      case WM_TOUCHUP:
      {
          int i = 0 + 1;
      }
      break;
  }

  return CallWindowProc(wndProc, hWnd, uMsg, wParam, lParam);
}

void
ClearSubclass(HWND hWnd)
{
  if (!hWnd)
  return;
  
  if (GetProp(hWnd, TEXT("MozillaWndProc"))) {
    SetWindowLongPtr(hWnd, GWLP_WNDPROC, (long)GetProp(hWnd, TEXT("MozillaWndProc")));
    RemoveProp(hWnd, TEXT("MozillaWndProc"));
    RemoveProp(hWnd, TEXT("MozillaWindow"));
  }
}

void
SetSubclass(HWND hWnd, NPWindow* window)
{
  // Subclass the plugin window so we can handle our own windows events.
  WNDPROC origProc = (WNDPROC)::SetWindowLong(hWnd, GWLP_WNDPROC, (long)PluginWndProc);
  SetProp(hWnd, TEXT("MozillaWndProc"), (HANDLE)origProc);
  SetProp(hWnd, TEXT("MozillaWindow"), (HANDLE)window);
}

NPError NPP_NewStream(NPP instance, NPMIMEType type, NPStream* stream, NPBool seekable, uint16_t* stype) {  return NPERR_GENERIC_ERROR; }
NPError NPP_DestroyStream(NPP instance, NPStream* stream, NPReason reason) { return NPERR_NO_ERROR; }
int32_t NPP_WriteReady(NPP instance, NPStream* stream) { return NPERR_NO_ERROR; }
int32_t NPP_Write(NPP instance, NPStream* stream, int32_t offset, int32_t len, void* buffer) { return NPERR_NO_ERROR; }
void NPP_StreamAsFile(NPP instance, NPStream* stream, const char* fname) {}
void NPP_Print(NPP instance, NPPrint* platformPrint) {}
int16_t NPP_HandleEvent(NPP instance, void* event) { return NPERR_NO_ERROR; }
void NPP_URLNotify(NPP instance, const char* url, NPReason reason, void* notifyData) {}
NPError NPP_GetValue(NPP instance, NPPVariable variable, void* value) { return NPERR_NO_ERROR; }
NPError NPP_SetValue(NPP instance, NPNVariable variable, void* value) { 
  if (variable != NPNVaudioDeviceChangeDetails)
    return NPERR_NO_ERROR; 

  NPAudioDeviceChangeDetails &details = *(NPAudioDeviceChangeDetails*)value;
  printf("NPAUDIOTEST: "
	  "Received Default Audio Device Change Notification: (%d,%d,%S)\n",
    details.flow, details.role, details.defaultDevice);

  return NPERR_NO_ERROR; 
}

